library(glmnet)
library(MASS)
NUM_MC = 2500;
N = 25;
M = 25;
dimX = 400;
K = 2;

################################################################################
# PARALLEL PROCESSING
################################################################################
library(foreach);
library(doParallel);
cl<-makeCluster(3);
registerDoParallel(cl);

################################################################
# REGRESSION PARAMETERS
################################################################
theta = 1.0;
pi1 = 1.0;
zeta = matrix( c( 0.5^(1:dimX) ), dimX, 1);
pi2  = matrix( c( 0.5^(1:dimX) ), dimX, 1);
xi   = matrix( c( 0.5^(1:dimX) ), dimX, 1);

################################################################
# PARAMETERS FOR GENERATION OF Xij, Xi, AND Xj
################################################################
muXij = matrix(0,dimX,1);
SigmaXij = matrix(0,dimX,dimX);
for(index1 in 1:dimX){
 for(index2 in 1:dimX){
  SigmaXij[index1,index2] = 0.25^abs(index1-index2);
 }
}
weightXij = 2/4;

muXi = matrix(0,dimX,1);
SigmaXi = matrix(0,dimX,dimX);
for(index1 in 1:dimX){
 for(index2 in 1:dimX){
  SigmaXi[index1,index2] = 0.25^abs(index1-index2);
 }
}
weightXi = 1/4;

muXj = matrix(0,dimX,1);
SigmaXj = matrix(0,dimX,dimX);
for(index1 in 1:dimX){
 for(index2 in 1:dimX){
  SigmaXj[index1,index2] = 0.25^abs(index1-index2);
 }
}
weightXj = 1/4;

################################################################
# PARAMETERS FOR GENERATION OF EPUPij, EPUPi, EPUPj
################################################################
muEPUPij = matrix(0,2,1);
SigmaEPUPij = matrix(c(1,0.25,0.25,1),2,2);
weightEPUPij = 2/4;

muEPUPi = matrix(0,2,1);
SigmaEPUPi = matrix(c(1,0.25,0.25,1),2,2);
weightEPUPi = 1/4;

muEPUPj = matrix(0,2,1);
SigmaEPUPj = matrix(c(1,0.25,0.25,1),2,2);
weightEPUPj = 1/4;

################################################################
# PARAMETERS FOR GENERATION OF Vij, Vi, AND Vj
################################################################
muVij = 0;
sigmaVij = 1;
weightVij = 2/4;

muVi = 0;
sigmaVi = 1;
weightVi = 1/4;

muVj = 0;
sigmaVj = 1;
weightVj = 1/4;





################################################################
# START MC
################################################################
LIST_RESULTS = NULL;
LIST_RESULTS <- foreach( iter = 1:NUM_MC, .combine = rbind )%dopar%{
res <- tryCatch({
library(glmnet)
library(MASS)

################################################################
# GENERATE DX FROM Xij, Xi, Xj
################################################################
aXij = mvrnorm(N*M, muXij, SigmaXij);

aXi = mvrnorm(N, muXi, SigmaXi);
aXi = aXi %x% matrix(1,M,1);

aXj = mvrnorm(M, muXj, SigmaXj);
aXj = matrix(1,N,1) %x% aXj;

X = weightXij * aXij + weightXi * aXi + weightXj * aXj;

################################################################
# GENERATE DX FROM EPUPij, EPUPi, EPUPj
################################################################
aEPUPij = mvrnorm(N*M, muEPUPij, SigmaEPUPij);

aEPUPi = mvrnorm(N, muEPUPi, SigmaEPUPi);
aEPUPi = aEPUPi %x% matrix(1,M,1);

aEPUPj = mvrnorm(M, muEPUPj, SigmaEPUPj);
aEPUPj = matrix(1,N,1) %x% aEPUPj;

EPUP = weightEPUPij * aEPUPij + weightEPUPi * aEPUPi + weightEPUPj * aEPUPj;
Epsilon = EPUP[,1];
Upsilon = EPUP[,2];

################################################################
# GENERATE E FROM Vij, Vi, Vj
################################################################
aVij = rnorm(N*M, muVij, sigmaVij);

aVi = mvrnorm(N, muVi, sigmaVi);
aVi = aVi %x% matrix(1,M,1);

aVj = mvrnorm(M, muVj, sigmaVj);
aVj = matrix(1,N,1) %x% aVj;

V = weightVij * aVij + weightVi * aVi + weightVj * aVj;

################################################################
# GENERATE Zij, Dij, AND Yij
################################################################
Z = X %*% xi + V;
D = Z %*% pi1 + X %*% pi2 + Upsilon;
Y = D %*% theta + X %*% zeta + Epsilon;





################################################################
# ESTIMATION
################################################################
list_idx = matrix(1:N,N,1) %x% matrix(1,M,1);
list_jdx = matrix(1,N,1) %x% matrix(1:M,M,1);
k = ceiling(list_idx / ceiling(N/K));
l = ceiling(list_jdx / ceiling(M/K));
Y_Xbeta = NULL;
D_Xgamma = NULL;
Z_Xxi = NULL;
subsample_mean_Y_Xbeta_Z_Xxi = NULL;
subsample_mean_D_Xgamma_Z_Xxi = NULL;
for(kdx in 1:max(k)){
 for(ldx in 1:max(l)){
  ##############################################################
  # GET AUXILIARY SAMPLE (I^c X J^c)
  ##############################################################
  auxiliary_indices = which( k != kdx & l != ldx );
  auxiliaryY = Y[auxiliary_indices];
  auxiliaryD = D[auxiliary_indices];
  auxiliaryZ = Z[auxiliary_indices];
  auxiliaryX = X[auxiliary_indices,];

  ##############################################################
  # LASSO OF Y ON X
  ##############################################################
  cvfit = cv.glmnet( auxiliaryX, auxiliaryY, standardize = TRUE, intercept = TRUE );
  mu1 = cvfit$lambda.min
  lasso1 = glmnet( auxiliaryX, auxiliaryY, lambda = mu1, standardize = TRUE, intercept = TRUE );
  beta_LASSO = coef( lasso1, s=mu1 );
  ##############################################################
  # LASSO OF D ON X
  ##############################################################
  cvfit = cv.glmnet( auxiliaryX, auxiliaryD, standardize = TRUE, intercept = TRUE );
  mu2 = cvfit$lambda.min
  lasso2 = glmnet( auxiliaryX, auxiliaryD, lambda = mu2, standardize = TRUE, intercept = TRUE );
  gamma_LASSO = coef( lasso2, s=mu2 );
  ##############################################################
  # LASSO OF Z ON X
  ##############################################################
  cvfit = cv.glmnet( auxiliaryX, auxiliaryZ, standardize = TRUE, intercept = TRUE );
  mu3 = cvfit$lambda.min
  lasso3 = glmnet( auxiliaryX, auxiliaryZ, lambda = mu3, standardize = TRUE, intercept = TRUE );
  xi_LASSO = coef( lasso3, s=mu3 );

  ##############################################################
  # GET MAIN SAMPLE (I X J)
  ##############################################################
  main_indices = which( k == kdx & l == ldx );
  mainY = Y[main_indices];
  mainD = D[main_indices];
  mainZ = Z[main_indices];
  mainX = matrix(X[main_indices,],length(main_indices),dim(X)[2]);

  ##############################################################
  # COMPUTE Y - Xbeta, D - Xgamma, AND Z - Xxi
  ##############################################################
  Y_Xbeta = rbind(Y_Xbeta, mainY - cbind(1,mainX) %*% beta_LASSO);
  D_Xgamma = rbind(D_Xgamma, mainD - cbind(1,mainX) %*% gamma_LASSO);
  Z_Xxi = rbind(Z_Xxi, mainZ - cbind(1,mainX) %*% xi_LASSO);

  ##############################################################
  # COMPUTE E(Y - Xbeta)*(Z - Xxi) AND E(D - Xgamma)*(Z - Xxi)
  ##############################################################
  subsample_mean_Y_Xbeta_Z_Xxi = rbind(subsample_mean_Y_Xbeta_Z_Xxi, mean((mainY - cbind(1,mainX) %*% beta_LASSO)*(mainZ - cbind(1,mainX) %*% xi_LASSO)));
  subsample_mean_D_Xgamma_Z_Xxi = rbind(subsample_mean_D_Xgamma_Z_Xxi, mean((mainD - cbind(1,mainX) %*% gamma_LASSO)*(mainZ - cbind(1,mainX) %*% xi_LASSO)));
 }
}

theta_hat = sum(subsample_mean_Y_Xbeta_Z_Xxi) / sum(subsample_mean_D_Xgamma_Z_Xxi);





################################################################
# VARIANCE ESTIMATION
################################################################
Epsilon_hat = Y_Xbeta - theta_hat * D_Xgamma;
Epsilon_hat = t(matrix(Epsilon_hat,N,M));
U_hat = D_Xgamma;
U_hat = t(matrix(D_Xgamma,N,M));
V_hat = Z_Xxi;
V_hat = t(matrix(V_hat,N,M));
Gamma = 0;
J = 0;
for(kdx in 1:max(k)){
 for(ldx in 1:max(l)){
  ##############################################################
  # GET MAIN SAMPLE (I X J)
  ##############################################################
  main_indices = which( k == kdx & l == ldx );
  curr_list_idx = unique(list_idx[main_indices]);
  curr_list_jdx = unique(list_jdx[main_indices]);
  CardinalityIk = length(curr_list_idx);
  CardinalityJl = length(curr_list_jdx);

  for(idx1 in curr_list_idx){
   for(jdx1 in curr_list_jdx){

    for(jdx2 in curr_list_jdx){
     cardinality = CardinalityIk^2 * CardinalityJl^2 / min(CardinalityIk,CardinalityJl);
     Gamma = Gamma + cardinality^(-1) * Epsilon_hat[idx1,jdx1] * V_hat[idx1,jdx1] * Epsilon_hat[idx1,jdx2] * V_hat[idx1,jdx2];
    }#END FOR jdx2
    for(idx2 in curr_list_idx){
     cardinality = CardinalityIk^2 * CardinalityJl^2 / min(CardinalityIk,CardinalityJl);
     Gamma = Gamma + cardinality^(-1) * Epsilon_hat[idx1,jdx1] * V_hat[idx1,jdx1] * Epsilon_hat[idx2,jdx1] * V_hat[idx2,jdx1];
    }#END FOR idx2

    J = J - U_hat[idx1,jdx1] * V_hat[idx1,jdx1] / (CardinalityIk * CardinalityJl);
   }#END FOR idx1
  }#END FOR jdx1


 }#END FOR kdx
}#END FOR ldx

Gamma = Gamma / K^2;
J = J / K^2;
sigma2 = Gamma / J^2;
theta_SD = sqrt( sigma2 / min(N,M) );

################################################################
# DISPLAY AND STORE RESULTS
################################################################
Sys.sleep(0.001);
print( c(iter, theta_hat, theta_SD, abs(theta_hat-theta) <= qnorm(0.975)*theta_SD) );
flush.console();
#LIST_RESULTS = rbind(LIST_RESULTS, c(theta_hat, theta_SD, abs(theta_hat-theta) <= qnorm(0.975)*theta_SD));

################################################################
# END MC
################################################################
c(theta_hat, theta_SD, abs(theta_hat-theta) <= qnorm(0.975)*theta_SD);
}, error=function(e) NULL)
}# END foreach iter #
stopCluster(cl);

bias = mean(LIST_RESULTS[,1]-theta);
SD = sd(LIST_RESULTS[,1]);
RMSE = (bias^2+SD^2)^0.5;
mean_SE = mean(LIST_RESULTS[,2]);
cover95 = mean(LIST_RESULTS[,3]);
c(bias,SD,RMSE,mean_SE,cover95);